/*
Converts a brainfuck file into brainloller code on a .png image.

Compiles with "g++ lodepng.cpp bftopng.cpp -o bftopng"
*/
#include "png.h"

#include <vector>

#include <iostream>
#include <fstream>
#include <cmath>
#include <string>


#include "lodepng.h"

using namespace std;

int main(int argc, char *argv[]) 
{
    if(!argv[1])
    {
        cout << "please specify a file\n";
        return 0;
    }

    std::vector<unsigned char> file;
    LodePNG::loadFile(file, argv[1]);
    std::string code;
    for(size_t i = 0; i < file.size(); i++)
    {
      char c = file[i];
      if(c == '<' || c == '>' || c == '+' || c == '-' || c == '[' || c == ']' || c == '.' || c == ',') code += c;
    }
    int size = code.size();
    if(size < 1) std::cout << "error in input file" << std::endl;
    
    //generate image data
    int iw = int(std::sqrt(size)) + 3; //extra pixels for the "turning around" to the next line
    int ih = int(std::sqrt(size)) + 1;
    std::vector<unsigned char> image(iw * ih * 3, 0);
    
    int x = 0, y = 0;
    
    for(int i = 0; i < size; i++)
    {
        bool nop = false;
        switch(code[i])
        {
            case '>':
                image[3 * iw * y + 3 * x + 0] = 255;
                image[3 * iw * y + 3 * x + 1] = 0;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            case '<':
                image[3 * iw * y + 3 * x + 0] = 128;
                image[3 * iw * y + 3 * x + 1] = 0;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            case '+':
                image[3 * iw * y + 3 * x + 0] = 0;
                image[3 * iw * y + 3 * x + 1] = 255;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            case '-':
                image[3 * iw * y + 3 * x + 0] = 0;
                image[3 * iw * y + 3 * x + 1] = 128;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            case '.':
                image[3 * iw * y + 3 * x + 0] = 0;
                image[3 * iw * y + 3 * x + 1] = 0;
                image[3 * iw * y + 3 * x + 2] = 255;
                break;
            case ',':
                image[3 * iw * y + 3 * x + 0] = 0;
                image[3 * iw * y + 3 * x + 1] = 0;
                image[3 * iw * y + 3 * x + 2] = 128;
                break;
            case '[':
                image[3 * iw * y + 3 * x + 0] = 255;
                image[3 * iw * y + 3 * x + 1] = 255;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            case ']':
                image[3 * iw * y + 3 * x + 0] = 128;
                image[3 * iw * y + 3 * x + 1] = 128;
                image[3 * iw * y + 3 * x + 2] = 0;
                break;
            default:
                nop = true;
                break;
        }
        
        if(!nop)
        {
            if(y % 2 == 0)
            {
                x++;
                if(x == iw - 1)
                {
                    //x++;
                    image[3 * iw * y + 3 * x + 0] = 0;
                    image[3 * iw * y + 3 * x + 1] = 255;
                    image[3 * iw * y + 3 * x + 2] = 255;
                    y++;
                    image[3 * iw * y + 3 * x + 0] = 0;
                    image[3 * iw * y + 3 * x + 1] = 255;
                    image[3 * iw * y + 3 * x + 2] = 255;
                    x--;
                }
            }
            else
            {
                x--;
                if(x == 0)
                {
                    //x--;
                    image[3 * iw * y + 3 * x + 0] = 0;
                    image[3 * iw * y + 3 * x + 1] = 128;
                    image[3 * iw * y + 3 * x + 2] = 128;
                    y++;
                    image[3 * iw * y + 3 * x + 0] = 0;
                    image[3 * iw * y + 3 * x + 1] = 128;
                    image[3 * iw * y + 3 * x + 2] = 128;
                    x++;
                }
            }
        }
    }
    
    LodePNG::encode("result.png", image, iw, ih, 2, 8);

    return 0;
}

/*
Copyright (c) 2005 by Lode Vandevenne.
All rights reserved.
*/
